#include <stdlib.h>
#include <errno.h>

#include "queue.h"

struct queue
   {
   void ** data;
   size_t max;
   size_t start;
   size_t count;
   };

////////////////////////////////////////////////////////////////////////////////
struct queue *
queue_new(size_t const max)
   {
   if (max < 1)
      return NULL;

   struct queue * queue = malloc(sizeof(struct queue));
   if (NULL == queue)
      return NULL;

   queue->start = queue->count = 0;
   queue->max = max;

   queue->data = calloc(queue->max, sizeof(void *));
   if (NULL == queue->data)
      {
      free(queue);
      return NULL;
      }

   return queue;
   }

////////////////////////////////////////////////////////////////////////////////
// Returns:
// NULL on error
// elem on success and the buffer is not full
// the overwritten element on success if the buffer was full
void *
queue_push(struct queue * queue, void * elem)
   {
   if ((NULL == queue) || (NULL == elem))
      return NULL;

   if (queue->count == queue->max)
      return NULL;

   size_t end = (queue->start + queue->count) % queue->max;

   queue->data[end] = elem;
   queue->count++;

   return elem;
   }

////////////////////////////////////////////////////////////////////////////////
void *
queue_pop(struct queue * queue)
   {
   if (NULL == queue)
      return NULL;

   if (0 == queue->count)
      return NULL;

   void * t = queue->data[queue->start];
   queue->count--;
   queue->start = (queue->start + 1) % queue->max;

   return t;
   }

////////////////////////////////////////////////////////////////////////////////
size_t
queue_size(struct queue * queue)
   {
   if (NULL == queue)
      return 0;

   return queue->count;
   }

////////////////////////////////////////////////////////////////////////////////
void
queue_delete(struct queue * queue)
   {
   if (NULL != queue) free(queue->data);
   free(queue);
   }
